 /**************************************************************************
 *
 * powertester.c
 *
 * (C) 2004, 2005 - Klaus S. Madsen
 */

//#define DEBUG_CRC

#include <sys/types.h>
#include <sys/socket.h>
#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>
#include <sys/select.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <time.h>
#include <sys/stat.h>

// For controlling the terminal
#include <termios.h>

// For ioctl
#include <sys/ioctl.h>

#include <list>

#include "crc.h"
#include "outdata.h"
#include <sys/time.h>
#include <time.h>

#ifdef DEBUG
#define DATADIR "./"
#else
#define DATADIR "/data/yada"
#endif

int packets;
int bytes = 0;
int recv_bytes;

struct timeval start_time;

void usage()
{
  printf("powertester - (C) 2004, 2005 Klaus S. Madsen\n\n");
  printf("usage:\n");
  printf("\npowertester\n\n");
  printf("Once the program is running, pressing CTRL+C will cause an exit\n");
}

char connected_addr[20];

// Go through the list of current connections. If one is for the
// requested device, return it's handle, otherwise create a new
// connection.
void connect_to_device(bdaddr_t *bda, int *dev_id, uint16_t *handle)
{
  // Figure out which device to use.
  *dev_id = hci_get_route(bda);

  // Open the device
  int fd = hci_open_dev(*dev_id);
  if (fd < 0) {
    perror("Error opening hci device");
    exit (1);
  }

  struct hci_conn_list_req *cl;
  
  cl = (hci_conn_list_req*)malloc(10 * sizeof(*cl->conn_info) + sizeof(*cl));
  if (!cl) {
    perror("Cannot allocate memory for connection list");
    exit(1);
  }

  cl->dev_id = *dev_id;
  cl->conn_num = 10;

  if (ioctl(fd, HCIGETCONNLIST, (void*)cl)) {
    perror("Cannot retrieve connection list");
    exit(1);
  }

  struct hci_conn_info *ci = cl->conn_info;
  for (int i = 0; i < cl->conn_num; i++, ci++) {
    if (!bacmp(bda, &ci->bdaddr)) {
      *handle = ci->handle;
      hci_close_dev(fd);
      return;
    }
  }

  uint8_t role = 0x01;
  int ptype = HCI_DH5;

  int res = hci_create_connection(fd, bda, htobs(ptype), 0, role, handle, 25000);
  if (res < 0) {
    hci_close_dev(fd);
    perror("Error connecting to device");
    exit(1);
  }

  hci_close_dev(fd);
}

// Open the HCI device, and make it ready for reading.
int open_hci_for_read(int dev_id)
{
  struct sockaddr_hci addr;
  int res = socket(AF_BLUETOOTH, SOCK_RAW, BTPROTO_HCI);
  if (res < 0) {
    perror("Error opening read socket");
    exit(1);
  }

  int opt = 1;
  if (setsockopt(res, SOL_HCI, HCI_DATA_DIR, &opt, sizeof(opt)) < 0) {
    perror("Can't enable data direction info");
    exit(1);
  }

  opt = 1;
  if (setsockopt(res, SOL_HCI, HCI_TIME_STAMP, &opt, sizeof(opt)) < 0) {
    perror("Can't enable time stamp");
    exit(1);
  }

  struct hci_filter flt;
  hci_filter_clear(&flt);
  hci_filter_all_ptypes(&flt);
  hci_filter_all_events(&flt);

  if (setsockopt(res, SOL_HCI, HCI_FILTER, &flt, sizeof(flt)) < 0) {
    perror("Can't set HCI filter");
    exit(1);
  }

  /* Bind socket to the HCI device */
  addr.hci_family = AF_BLUETOOTH;
  addr.hci_dev = dev_id;
  if (bind(res, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
    printf("Can't attach to device hci%d. %s(%d)\n", 
	   dev_id, strerror(errno), errno);
    exit(1);
  }


  // The socket is now ready for reading
  return res;
}

int should_exit;

void connection_end()
{
  struct timeval end_time;

  gettimeofday(&end_time, 0);

  double seconds = (end_time.tv_sec - start_time.tv_sec) +
    (end_time.tv_usec - start_time.tv_usec) / 1000000.0;
  fprintf(stderr, "\n");
  printf("\nPackets: %d\nBytes: %d\nTime: %f\n"
	 "Rate: %f\n", 
	 packets - 1, bytes, seconds, bytes / seconds);
}

void handle_bluetooth(int read_fd, int handle)
{
  struct cmsghdr *cmsg;
  struct msghdr msg;
  struct iovec iv;
  int direction;

  char buf[1024];
  char ctrl[1024];

  iv.iov_base = buf;
  iv.iov_len = 1024;

  memset(&msg, 0, sizeof(msg));

  msg.msg_iov = &iv;
  msg.msg_iovlen = 1;
  msg.msg_control = ctrl;
  msg.msg_controllen = 1024;

  ssize_t res = recvmsg(read_fd, &msg, 0);
  if (res < 0) {
    perror("Failure reading from bluetooth");
    exit(1);
  }

  cmsg = CMSG_FIRSTHDR(&msg);
  while (cmsg) {
    switch (cmsg->cmsg_type) {
    case HCI_CMSG_DIR:
      direction = *((int *)CMSG_DATA(cmsg));
      break;
    }
    cmsg = CMSG_NXTHDR(&msg, cmsg);
  }

  // Skip all outgoing packets.
  if (direction != 1)
    return;

  if (buf[0] == HCI_ACLDATA_PKT) {
    hci_acl_hdr *hdr = (hci_acl_hdr*) &buf[1];
    if (acl_handle(hdr->handle) == handle) {
      //      printf("Package size: %d\n", hdr->dlen);
      if (packets == 1) 
	gettimeofday(&start_time, 0);

      if (packets >= 1) 
	bytes += hdr->dlen;

      packets++;  
    
      fprintf(stderr, ".");
      //      printf("%d, %d\n", bytes, hdr->dlen);

      if (bytes == recv_bytes) {
	connection_end();
	exit(0);
      }
    }
  } else if (buf[0] == HCI_EVENT_PKT) {
    hci_event_hdr *hdr = (hci_event_hdr*) &buf[1];
    if (hdr->evt == EVT_DISCONN_COMPLETE) {
      evt_disconn_complete *discon 
	= (evt_disconn_complete*)(buf + 1 + sizeof(hci_event_hdr));
      if (discon->handle == handle) {
	fprintf(stderr, "Disconnect recieved.\n");
	connection_end();
	exit(1);
      }
    }
  }
  fflush(stdout);
}

int read_fd;

void retrieve_data_from_device(bdaddr_t *bda)
{
  char address_str[20];
  ba2str(bda, address_str);
  
  uint16_t handle;
  int dev_id;

  // Figure out which device, and which handle to use for
  // communicating with the client.  

  // This should be split up into two functions. One to get the
  // dev_id, and one to connect. That way we can setup everything
  // before we connect. Otherwise we might miss data.

  connect_to_device(bda, &dev_id, &handle);

  if (!read_fd)
    read_fd = open_hci_for_read(dev_id);

  fprintf(stderr, "Connected.\n");

  timeval time1;
  timeval time2;

  packets = 0;
  should_exit = 0;

  gettimeofday(&time1, 0);

  fd_set fds;
  do {
    FD_ZERO(&fds);
    FD_SET(read_fd, &fds);

    // Wait for input on either stdin or the read_fd.
    int res = select(read_fd + 1, &fds, 0, 0, 0);
    if (res == -1) {
      // An interrupt will happen when the user presses CTRL+C
      if (errno == EINTR)
	continue;

      perror("Error selecting");
      exit(1);
    } else {
      // We have a socket ready for reading.
      handle_bluetooth(read_fd, handle);
    }
  } while (!should_exit);
}


std::list<bdaddr_t> addresses;

void setup_addresses()
{
  int i;
  bdaddr_t addrs[12] = { { 0x08, 0x4d, 0x17, 0x37, 0x80, 0x00 },
			 { 0x09, 0x4d, 0x17, 0x37, 0x80, 0x00 },
			 { 0x10, 0x4d, 0x17, 0x37, 0x80, 0x00 },
			 { 0x0b, 0x4d, 0x17, 0x37, 0x80, 0x00 },
			 { 0x2a, 0x4d, 0x17, 0x37, 0x80, 0x00 },
			 { 0x2f, 0x4d, 0x17, 0x37, 0x80, 0x00 },
			 { 0x3b, 0x4d, 0x17, 0x37, 0x80, 0x00 },
			 { 0xbb, 0x4e, 0x17, 0x37, 0x80, 0x00 },
			 { 0xbc, 0x4e, 0x17, 0x37, 0x80, 0x00 },
			 { 0xbd, 0x4e, 0x17, 0x37, 0x80, 0x00 },
			 { 0xbe, 0x4e, 0x17, 0x37, 0x80, 0x00 },
			 { 0xbf, 0x4e, 0x17, 0x37, 0x80, 0x00 } };
  
  for (i = 0; i < 12; i++)
    addresses.push_back(addrs[i]);

}

int main(int argc, char *argv[])
{
  srand(time(0));

  if (argc != 2) {
    // If there are arguments, print usage, and exit
    usage();
    exit(1);
  }

  recv_bytes = atoi(argv[1]);

  setup_addresses();

  read_fd = 0;

  bool cont = true;

  while (cont) {
    inquiry_info *info = 0;
    fprintf(stderr, "Starting scan\n");
    int no_devs = hci_inquiry(-1, 8, 0, 0, &info, 0);
    if (no_devs < 0) {
      perror("Inquery Failed");
      exit(1);
    }

    for (int i = 0; i < no_devs; i++) {
      char addr[20];
      bool matches = false;
      ba2str(&info[i].bdaddr, addr);
      for (std::list<bdaddr_t>::iterator cur_adr = addresses.begin();
	   cur_adr != addresses.end(); ++cur_adr) {
	if (bacmp(&(*cur_adr), &info[i].bdaddr) == 0) {
	  addresses.push_back(*cur_adr);
	  addresses.erase(cur_adr);
	  retrieve_data_from_device(&info[i].bdaddr);
	  matches = true;
	  break;
	}
      }
      if (matches) {
	// We don't want to look at the rest of the elements, as they
	// might already have been handled.
	cont = false;
	break; 
      }
    }
  }
}
