 /**************************************************************************
 *
 * powertester.c
 *
 * (C) 2004, 2005 - Klaus S. Madsen
 */

//#define DEBUG_CRC

#include <sys/types.h>
#include <sys/socket.h>
#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>
#include <sys/select.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <time.h>
#include <sys/stat.h>

// For controlling the terminal
#include <termios.h>

// For ioctl
#include <sys/ioctl.h>

#include <list>

#include "crc.h"
#include "outdata.h"

#ifdef DEBUG
#define DATADIR "./"
#else
#define DATADIR "/data/yada"
#endif

enum protocol_state_t {
  PS_Searching,
  PS_Connected,
  PS_WaitingForData,
};

protocol_state_t state;

// We know that there can never be more than 128KB/256B = 512 pages
// that we should recieve.
received_pages_t pages[512];

uint16_t init_data_count;
struct init_data_t *init_data = 0;

int packets;
int bytes;

void usage()
{
  printf("powertester - (C) 2004, 2005 Klaus S. Madsen\n\n");
  printf("usage:\n");
  printf("\npowertester\n\n");
  printf("Once the program is running, pressing CTRL+C will cause an exit\n");
}

char connected_addr[20];

void create_node_dir(bdaddr_t *bda)
{
  char buffer[256];
  ba2str(bda, connected_addr);

  snprintf(buffer, 256, "%s/%s", DATADIR, connected_addr);
  mkdir(buffer, S_IRWXU | S_IRWXG | S_IRWXO);
}

// Go through the list of current connections. If one is for the
// requested device, return it's handle, otherwise create a new
// connection.
void connect_to_device(bdaddr_t *bda, int *dev_id, uint16_t *handle)
{
  // Figure out which device to use.
  *dev_id = hci_get_route(bda);

  // Open the device
  int fd = hci_open_dev(*dev_id);
  if (fd < 0) {
    perror("Error opening hci device");
    exit (1);
  }

  struct hci_conn_list_req *cl;
  
  cl = (hci_conn_list_req*)malloc(10 * sizeof(*cl->conn_info) + sizeof(*cl));
  if (!cl) {
    perror("Cannot allocate memory for connection list");
    exit(1);
  }

  cl->dev_id = *dev_id;
  cl->conn_num = 10;

  if (ioctl(fd, HCIGETCONNLIST, (void*)cl)) {
    perror("Cannot retrieve connection list");
    exit(1);
  }

  struct hci_conn_info *ci = cl->conn_info;
  for (int i = 0; i < cl->conn_num; i++, ci++) {
    if (!bacmp(bda, &ci->bdaddr)) {
      *handle = ci->handle;
      hci_close_dev(fd);
      return;
    }
  }

  uint8_t role = 0x01;
  int ptype = HCI_DH5;

  int res = hci_create_connection(fd, bda, htobs(ptype), 0, role, handle, 25000);
  if (res < 0) {
    hci_close_dev(fd);
    perror("Error connecting to device");
    exit(1);
  }

  hci_close_dev(fd);
  create_node_dir(bda);
  state = PS_Connected;
  
}

// Open the HCI device, and make it ready for reading.
int open_hci_for_read(int dev_id)
{
  int res = hci_open_dev(dev_id);
  if (res < 0) {
    perror("Error opening read socket");
    exit(1);
  }

  struct hci_filter flt;
  hci_filter_clear(&flt);
  hci_filter_all_ptypes(&flt);
  hci_filter_all_events(&flt);

  if (setsockopt(res, SOL_HCI, HCI_FILTER, &flt, sizeof(flt)) < 0) {
    perror("Can't set HCI filter");
    exit(1);
  }

  int opt = 1;
  if (setsockopt(res, SOL_HCI, HCI_DATA_DIR, &opt, sizeof(opt)) < 0) {
    perror("Can't enable data direction info");
    exit(1);
  }

  // The socket is now ready for reading
  return res;
}

int should_exit;

int send_acl_data(int fd, uint16_t handle, const char *data, size_t len)
{
  uint8_t type = HCI_ACLDATA_PKT;
  hci_acl_hdr hc;
  struct iovec iv[4];
  int ivn;
  uint16_t crc = 0xFFFF;

  for (size_t i = 0; i < len; i++)
    crc = crcByte(crc, (uint8_t)data[i]);

  printf("CRC is 0x%04x\n", crc);

  hc.handle = htobs(acl_handle_pack(handle, 0));
  hc.dlen = len + sizeof(crc);

  printf("dlen is 0x%04x\n", hc.dlen);

  iv[0].iov_base = &type;
  iv[0].iov_len  = 1;
  iv[1].iov_base = &hc;
  iv[1].iov_len  = HCI_ACL_HDR_SIZE;
  iv[2].iov_base = (void*)data;
  iv[2].iov_len  = len;
  iv[3].iov_base = &crc;
  iv[3].iov_len = sizeof(crc);
  ivn = 4;

  while (writev(fd, iv, ivn) < 0) {
    if (errno == EAGAIN || errno == EINTR)
      continue;
    return -1;
  }
  
  return 0;
}

bool pages_init = false;

void save_data()
{
  FILE *f;
  char buf[100];

  if (init_data) {
    snprintf(buf, 100, "%s/%s/%08x", DATADIR, connected_addr,
	     init_data->init_data.first_page_time);

    f = fopen(buf, "w");
    if (f) {
      uint16_t highest_page = 0;
      
      for (uint16_t i = 0; i < 512; i++) {
	if (pages[i].received)
	  highest_page = i;
      }
      
      fwrite(&init_data_count, sizeof(init_data_count), 1, f);
      fwrite(&highest_page, sizeof(highest_page), 1, f);
      
      fwrite(init_data, sizeof(*init_data), init_data_count, f);
      fwrite(pages, sizeof(pages[0]), highest_page + 1, f);
      printf("Saved data: %d init data structures, and %d pages\n",
	     init_data_count, highest_page + 1);
      
      fclose(f);
    }
  }
}

int send_page_req(int fd, uint16_t handle, struct n2s_init *init)
{
  struct s2n_page_req prq[64];
  int prq_ptr = 0;

  struct timeval tv;

  gettimeofday(&tv, 0);
  
  if (!pages_init) {
    FILE *f;
    char buf[100];

    pages_init = true;

    snprintf(buf, 100, "%s/%s/%08x", DATADIR, connected_addr, 
	     init->first_page_time);
    
    f = fopen(buf, "r");
    if (f) {
      uint16_t highest_page;

      printf("Reading data from file\n");
      fread(&init_data_count, sizeof(init_data_count), 1, f);
      fread(&highest_page, sizeof(highest_page), 1, f);

      init_data_count += 1;

      // Allocate room for all the init-data.
      init_data = new init_data_t[init_data_count];

      fread(init_data, sizeof(*init_data), init_data_count - 1, f);
      fread(pages, sizeof(pages[0]), highest_page + 1, f);
      fclose(f);
    } else {
      printf("No old data found\n");
      init_data_count = 1;
      init_data = new init_data_t[1];
    }

    memcpy(&init_data[init_data_count - 1].init_data, init, sizeof(*init));
    memcpy(&init_data[init_data_count - 1].receive_time, &tv, sizeof(tv));
  }

  int16_t first_missing = -1;

  for (int i = 0 ; i < init->data_pages_used && prq_ptr < 64; i++) {
    if (pages[i].received) { 
      if (first_missing != -1) {
	prq[prq_ptr].first_page = first_missing;
	prq[prq_ptr].last_page = i - 1;
	prq_ptr++;
	first_missing = -1;
      }
    } else {
      if (first_missing == -1)
	first_missing = i;
    }
  }

  if (first_missing != -1) {
    prq[prq_ptr].first_page = first_missing;
    prq[prq_ptr].last_page = init->data_pages_used - 1;
    prq_ptr++;
  }

  printf("Requesting pages:\n");
  for (int i = 0; i < prq_ptr; i++) 
    printf("  %d - %d\n", prq[i].first_page, prq[i].last_page);

  prq[prq_ptr].first_page = -1;
  prq[prq_ptr].last_page = prq_ptr ? -1 : init->data_pages_used - 1;

  int res = send_acl_data(fd, handle, (char *)prq, 
			  sizeof(struct s2n_page_req) * (prq_ptr + 1));
  if (!res) 
    state = PS_WaitingForData;

  if (!prq_ptr) {
    should_exit = true;
    printf("Got last packet. %d packets in total\n", packets);
    save_data();
  }

  return res;
}

bool check_crc(char *buf)
{
  hci_acl_hdr *hdr = (hci_acl_hdr*) &buf[1];
  uint16_t crc = 0xFFFF;

  for (int i = 0; i < hdr->dlen - 2; i++)
    crc = crcByte(crc, (uint8_t)buf[sizeof(*hdr) + 1 + i]);

  printf("CRC: 0x%04x\n", crc);
  printf("CRC in packet: 0x%04x\n", *(uint16_t*)(buf + 1 + sizeof(*hdr) 
					 + hdr->dlen - 2));

#ifdef DEBUG_CRC
  if (hdr->dlen > 200) {
    int r = rand();
    if (r > RAND_MAX / 4)
      return false;
  }
#endif

  return crc == *(uint16_t*)(buf + 1 + sizeof(*hdr) + hdr->dlen - 2);
}

void handle_bluetooth(int read_fd, int handle, int write_fd)
{
  struct cmsghdr *cmsg;
  struct msghdr msg;
  struct iovec iv;
  int direction;

  char buf[1024];
  char ctrl[100];

  iv.iov_base = buf;
  iv.iov_len = 1024;

  msg.msg_iov = &iv;
  msg.msg_iovlen = 1;
  msg.msg_control = ctrl;
  msg.msg_controllen = 100;

  ssize_t res = recvmsg(read_fd, &msg, 0);
  if (res < 0) {
    perror("Failure reading from bluetooth");
    return;
  }

  cmsg = CMSG_FIRSTHDR(&msg);
  while (cmsg) {
    switch (cmsg->cmsg_type) {
    case HCI_CMSG_DIR:
      direction = *((int *)CMSG_DATA(cmsg));
      break;
    }
    cmsg = CMSG_NXTHDR(&msg, cmsg);
  }

  // Skip all outgoing packets.
  if (direction != 1)
    return;

  if (buf[0] == HCI_ACLDATA_PKT) {
    hci_acl_hdr *hdr = (hci_acl_hdr*) &buf[1];
    if (acl_handle(hdr->handle) == handle) {
      printf("Package size: %d\n", hdr->dlen);
      
      packets++;
      bytes += hdr->dlen;

      if (check_crc(buf)) {
	printf("CRC match\n");
	switch (state) {
	case PS_Connected: 
	  {
	    struct n2s_init * init = (struct n2s_init*)
	      (buf + 1 + sizeof(*hdr));

	    printf("Time\t\t\t\t: %u\n", init->time);
	    printf("Timestamp of first packet\t: %u\n", 
		   init->first_page_time);
	    printf("Digi data points\t\t: %u\n", init->digi_data_points);
	    printf("Analog data points\t\t: %u\n", init->analog_data_points);
	    printf("Put failures\t\t\t: %u\n", init->put_failures);
	    printf("Get failures\t\t\t: %u\n", init->get_failures);
	    printf("Bluetooth errors\t\t: %u\n", init->bluetooth_errors);
	    printf("WP failures\t\t\t: %u\n", init->write_page_failures);
	    printf("Data pages in use\t\t: %u\n", init->data_pages_used);
	    printf("Number of reboots\t\t: %u\n", init->reboots);
	    printf("Current power level\t\t: %u\n", init->power);
	    if (send_page_req(write_fd, handle, init)) {
	      perror("Error sending page request");
	      exit(0);
	    }
	  }
	  break;
	case PS_WaitingForData:
	  {
	    struct n2s_page * page = (struct n2s_page*)
	      (buf + sizeof(*hdr) + 1);

	    if (hdr->dlen == 4 || page->page == -1) {
	      send_page_req(write_fd, handle, 
			    &init_data[init_data_count - 1].init_data);
	      return;
	    }

	    for (int i = 0; i < hdr->dlen; i++) {
	      printf("%02x ", (uint8_t)buf[sizeof(*hdr) + 1 + i]);
	      if (i % 16 == 15)
		printf("\n");
	    }
	    printf("\n");

	    pages[page->page].received = true;
	    memcpy(pages[page->page].data, page->data, SPM_PAGESIZE);
	  }
	  break;
	case PS_Searching:
	  {
	    fprintf(stderr, "Wrong state in handle_bluetooth\n");
	    exit(1);
	  }
	}
      } else {
	printf("!!CRC does not match!!\n");
      }
    }
  } else if (buf[0] == HCI_EVENT_PKT) {
    hci_event_hdr *hdr = (hci_event_hdr*) &buf[1];
    if (hdr->evt == EVT_DISCONN_COMPLETE) {
      evt_disconn_complete *discon 
	= (evt_disconn_complete*)(buf + 1 + sizeof(hci_event_hdr));
      if (discon->handle == handle) {
	save_data();
	printf("Disconnect recieved.\n");
	exit(1);
      }
    }
  }
  fflush(stdout);
}

int read_fd;

void retrieve_data_from_device(bdaddr_t *bda)
{
  char address_str[20];
  ba2str(bda, address_str);
  
  printf("Will connect to %s\n", address_str);
  fflush(stdout);
  uint16_t handle;
  int dev_id;

  // Figure out which device, and which handle to use for
  // communicating with the client.  

  // This should be split up into two functions. One to get the
  // dev_id, and one to connect. That way we can setup everything
  // before we connect. Otherwise we might miss data.
  connect_to_device(bda, &dev_id, &handle);

  if (!read_fd)
    read_fd = open_hci_for_read(dev_id);
  int write_fd = hci_open_dev(dev_id);

  printf("Connected.\n");
  fflush(stdout);

  timeval time1;
  timeval time2;

  packets = 0;
  should_exit = 0;

  gettimeofday(&time1, 0);

  fd_set fds;
  do {
    FD_ZERO(&fds);
    FD_SET(read_fd, &fds);

    // Wait for input on either stdin or the read_fd.
    int res = select(read_fd + 1, &fds, 0, 0, 0);
    if (res == -1) {
      // An interrupt will happen when the user presses CTRL+C
      if (errno == EINTR)
	continue;

      perror("Error selecting");
      exit(1);
    } else {
      // We have a socket ready for reading.
      handle_bluetooth(read_fd, handle, write_fd);
    }
  } while (!should_exit);

  gettimeofday(&time2, 0);

  double rate;
  double used_time = (time2.tv_sec + time2.tv_usec / 1000000.0) 
    - (time1.tv_sec + time1.tv_usec / 1000000.0);

  rate = bytes / used_time;

  printf("%d bytes in %4.2f seconds (", bytes, used_time);
  
  if (rate < 512)
    printf("%4.2f ", rate);
  else 
    printf("%4.2f k", rate / 1024);

  printf("bytes/s)\n");
  fflush(stdout);

  if (hci_disconnect(write_fd, handle, HCI_OE_USER_ENDED_CONNECTION, 100) < 0)
    perror("Disconnect failed");

  hci_close_dev(write_fd);
}

std::list<bdaddr_t> addresses;

void setup_addresses()
{
  int i;
  bdaddr_t addrs[12] = { { 0x08, 0x4d, 0x17, 0x37, 0x80, 0x00 },
			 { 0x09, 0x4d, 0x17, 0x37, 0x80, 0x00 },
			 { 0x10, 0x4d, 0x17, 0x37, 0x80, 0x00 },
			 { 0x0b, 0x4d, 0x17, 0x37, 0x80, 0x00 },
			 { 0x2a, 0x4d, 0x17, 0x37, 0x80, 0x00 },
			 { 0x2f, 0x4d, 0x17, 0x37, 0x80, 0x00 },
			 { 0x3b, 0x4d, 0x17, 0x37, 0x80, 0x00 },
			 { 0xbb, 0x4e, 0x17, 0x37, 0x80, 0x00 },
			 { 0xbc, 0x4e, 0x17, 0x37, 0x80, 0x00 },
			 { 0xbd, 0x4e, 0x17, 0x37, 0x80, 0x00 },
			 { 0xbe, 0x4e, 0x17, 0x37, 0x80, 0x00 },
			 { 0xbf, 0x4e, 0x17, 0x37, 0x80, 0x00 } };
  
  for (i = 0; i < 12; i++)
    addresses.push_back(addrs[i]);

}

int main(int argc, char *argv[])
{
  srand(time(0));

  if (argc != 1) {
    // If there are arguments, print usage, and exit
    usage();
    exit(1);
  }

  for (int i = 0; i < 512; i++)
    pages[i].received = false;

  setup_addresses();

  read_fd = 0;

  bool cont = true;
  state = PS_Searching;

  while (cont) {
    inquiry_info *info = 0;
    printf("Starting scan\n");
    fflush(stdout);
    int no_devs = hci_inquiry(-1, 8, 0, 0, &info, IREQ_CACHE_FLUSH);
    if (no_devs < 0) {
      perror("Inquery Failed");
      exit(1);
    }
    printf("Scan finished, %d devices found\n", no_devs);
    fflush(stdout);

    for (int i = 0; i < no_devs; i++) {
      char addr[20];
      bool matches = false;
      ba2str(&info[i].bdaddr, addr);
      printf("Checking device with mac: %s\n", addr);
      fflush(stdout);
      for (std::list<bdaddr_t>::iterator cur_adr = addresses.begin();
	   cur_adr != addresses.end(); ++cur_adr) {
	if (bacmp(&(*cur_adr), &info[i].bdaddr) == 0) {
	  addresses.push_back(*cur_adr);
	  addresses.erase(cur_adr);
	  retrieve_data_from_device(&info[i].bdaddr);
	  matches = true;
	  break;
	}
      }
      if (matches) {
	// We don't want to look at the rest of the elements, as they
	// might already have been handled.
	cont = false;
	break; 
      }
    }
  }
}
