#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <time.h>

#define BUFFER_SIZE 4096
#define STDOUT 1
#define STDERR 2

#ifdef DEBUG
#define POWERTESTER_PATH "./powertester"
#define LOGFILE_PATH "./"
#else
#define POWERTESTER_PATH "/usr/local/bin/powertester"
#define LOGFILE_PATH "/var/log/powertester/"
#endif

int daemonize = 1;

// Write all the buffer contents before returning... No matter how
// many signals are recieved.
void write_all(int fd, const char *buffer, ssize_t bytes)
{
  ssize_t written = 0;
      
  do {
    ssize_t tmp = write(fd, buffer + written, bytes - written);
    if (tmp == -1  && errno != EINTR) {
      perror("Error writing");
      exit(1);
    }
    if (tmp > 0)
      written += tmp;
    
  } while (written < bytes);
}

/* Find the first free filename */
int find_free_logfile(const char* extension)
{
  int number = 0;
  int file_exists = 1;
  int res;
  time_t t;
  char *time_str;

  while (file_exists) {
    char name[100];
    snprintf(name, 100, LOGFILE_PATH "powertester%04d.%s", number, extension);

    // Try to open the file. If it fails the file exists.
    res = open(name, O_WRONLY | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
    file_exists = res < 0;
    number++;
  }

  t = time(0);
  time_str = ctime(&t);

  // Change the newline to a colon
  *(time_str + strlen(time_str) - 1) = ':';

  write_all(res, time_str, strlen(time_str));
  write_all(res, " Opened logfile\n", strlen(" Opened logfile\n"));

  return res; 
}

int should_exit = 0;

void sig_handler(int signal)
{
  should_exit = 1;
}

void setup_signal_handler()
{
  struct sigaction sa;

  memset(&sa, 0, sizeof(sa));
  sa.sa_handler = sig_handler;
  sa.sa_flags = SA_ONESHOT;
  
  if (sigaction(SIGINT, &sa, 0) < 0) {
    perror("Error enabling the signal handler");
    exit(1);
  }

  if (sigaction(SIGTERM, &sa, 0) < 0) {
    perror("Error enabling the signal handler");
    exit(1);
  }

}

void write_line(int fd, const char *start, int length)
{
  time_t t;
  char* time_str;

  t = time(0);
  time_str = ctime(&t);
  *(time_str + strlen(time_str) - 1) = ':';
  write_all(fd, time_str, strlen(time_str));
  write_all(fd, " ", 1);
  write_all(fd, start, length);
}

void read_and_output(int infd, int outfd, char *buf_start, char** buf_pos, int *include)
{
  char *nl_pos;
  int bytes_read = read(infd, *buf_pos, BUFFER_SIZE - (*buf_pos - buf_start));

  if (bytes_read == -1 && errno != EINTR) {
    perror("Error reading from child process");
    exit(1);
  }

  if (bytes_read <= 0) {
    *include = 0;
    return;
  }
      
  *buf_pos += bytes_read;
  **buf_pos = '\0';
  while ((nl_pos = strchr(buf_start, '\n'))) {
    // If we have found a new-line thus far.
    if (!daemonize) {
      // Write the line to STDOUT.
      write_all(STDOUT, buf_start, nl_pos - buf_start + 1);
    }
	
    // Write the line + the space in the beginning to the logfile.
    write_line(outfd, buf_start, nl_pos - buf_start + 1);
	
    // Move the rest of the string to the beginning of the buffer
    // (keeping the space)
    memmove(buf_start, nl_pos + 1, *buf_pos - nl_pos);
    *buf_pos -= nl_pos - buf_start + 1;
  }
}

int main(int argc, char *argv[]) 
{
  char stdout_buf[BUFFER_SIZE];
  char stderr_buf[BUFFER_SIZE];
  char *stdout_pos;
  char *stderr_pos;
  pid_t child_pid;
  int stdoutfds[2];
  int stderrfds[2];
  int status;
  int logfile;
  int errfile;
  int flags;
  int include_stderr;
  int include_stdout;

  setup_signal_handler();

  if (argc > 1 && strcmp(argv[1], "-d") == 0) {
    daemonize = 0;
  }

  if (daemonize) {
    if (fork())
      exit(0);

    setsid();
    fclose(stdin);
    fclose(stdout);
    fclose(stderr);
  }

  logfile = find_free_logfile("log");
  errfile = find_free_logfile("err");

  while (!should_exit) {
    /* Prepare a pipe, and fork the process. */
    pipe(stdoutfds);
    pipe(stderrfds);

    child_pid = fork();
    if (!child_pid) {
      // I am the child.
      // Close all unecessary filedescriptors.
      if (!daemonize) {
	close(STDOUT);
	close(STDERR);
      }
      close(stdoutfds[0]);
      close(stderrfds[0]);
      close(logfile);
      close(errfile);
      
      // Make the write end of the pipe STDOUT and STDERR
      dup2(stdoutfds[1], STDOUT);
      dup2(stderrfds[1], STDERR);

      // Close the endpoint, now that it have been copied to STDOUT
      // and STDERR.
      close(stdoutfds[1]);
      close(stderrfds[1]);
      
      // Execute the powertester-application
      execlp(POWERTESTER_PATH, "powertester", (char*)0);
      perror("Error executing " POWERTESTER_PATH);
    }
    close(stdoutfds[1]);
    close(stderrfds[1]);

    flags = fcntl(stdoutfds[0], F_GETFL);
    fcntl(stdoutfds[0], F_SETFL, flags | O_NONBLOCK);
    
    flags = fcntl(stderrfds[0], F_GETFL);
    fcntl(stderrfds[0], F_SETFL, flags | O_NONBLOCK);

    include_stderr = 1;
    include_stdout = 1;

    // Reset the buffer position
    stdout_pos = stdout_buf;
    stderr_pos = stderr_buf;
    
    do {
      fd_set fds;
      int res;

      if (should_exit) {
	const char *killmsg = "Got exit signal, killing child\n";
	write_line(logfile, killmsg, strlen(killmsg));
	write_line(errfile, killmsg, strlen(killmsg));
	if (!daemonize) 
	  write_all(STDOUT, killmsg, strlen(killmsg));

	kill(child_pid, SIGKILL);
	break;
      }

      FD_ZERO(&fds);

      if (include_stdout)
	FD_SET(stdoutfds[0], &fds);

      if (include_stderr)
	FD_SET(stderrfds[0], &fds);

      res = select(stdoutfds[0] > stderrfds[0] ? stdoutfds[0] + 1 : stderrfds[0] + 1,
		   &fds, 0, 0, 0);
      
      if (res < 0) {
	char buf[1024];
	snprintf(buf, 1024, "Error selecting: %s. Will kill child next.\n", 
		 strerror(errno));
	write_line(errfile, buf, strlen(buf));
	write_line(logfile, buf, strlen(buf));
	if (!daemonize)
	  write_all(STDOUT, buf, strlen(buf));
		    
	should_exit = 1;
      }

      if (FD_ISSET(stdoutfds[0], &fds)) {
	read_and_output(stdoutfds[0], logfile, stdout_buf, &stdout_pos, 
			&include_stdout);
      } else if (FD_ISSET(stderrfds[0], &fds)) {
	read_and_output(stderrfds[0], errfile, stderr_buf, &stderr_pos,
			&include_stderr);
      }

    } while (include_stdout || include_stderr);

    if (stdout_buf != stdout_pos) {
      if (!daemonize) {
	write_all(STDOUT, stdout_buf, stdout_pos - stdout_buf + 1);
      }
	
      write_line(logfile, stdout_buf, stdout_pos - stdout_buf + 1);
    }

    if (stderr_buf != stderr_pos) {
      if (!daemonize) {
	write_all(STDOUT, stderr_buf, stderr_pos - stderr_buf + 1);
      }
	
      write_line(errfile, stderr_buf, stderr_pos - stderr_buf + 1);
    }

    waitpid(child_pid, &status, 0);

    {
      char tmp_buf[200];
      snprintf(tmp_buf, 200, 
	       "Child exited with status: %d, and signal: %d\n", 
	       WEXITSTATUS(status), WTERMSIG(status));

      write_line(logfile, tmp_buf, strlen(tmp_buf));

      if (WEXITSTATUS(status) || WTERMSIG(status))
	write_line(errfile, tmp_buf, strlen(tmp_buf));

      if (!daemonize) {
	write_all(STDOUT, tmp_buf, strlen(tmp_buf));
      }
    }
      
    fflush(stdout);
    close(stdoutfds[0]);
    close(stderrfds[0]);
  }

  return 0;
}
